#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/select.h>
#include <unistd.h>
#include <cairo/cairo.h>
#include <cairo/cairo-xlib.h>
#include <cairo/cairo-pdf.h>
#include <string.h>

#define WIDTH (15*42+10)
#define HEIGHT (15*42+30+5+50+30)
float SCALE = 0.5;

Display *d;
Window w;
Pixmap p;
cairo_surface_t *cs;
cairo_surface_t *csout;
cairo_surface_t *letters;

int myscore, otherscore;
char hand[100];
int remain;
int points;
char game[2048];
char played[2048];
char other[256];
char winner[256];
int meplayed;
int starx, stary;
int pos, npos;

char *init_board[15] = {
        "..T..d...d..T..",
        ".t....d.d....t.",
        "T......D......T",
        "...D.......D...",
        "....t.....t....",
        "d....d...d....d",
        ".d....d.d....d.",
        "..D.........D..",
        ".d....d.d....d.",
        "d....d...d....d",
        "....t.....t....",
        "...D.......D...",
        "T......D......T",
        ".t....d.d....t.",
        "..T..d...d..T.."
};

int utf8(unsigned char *n, int *_len)
{
  int v = 0;
  if (*n == 0xc0 || *n == 0xc1 ||
      *n >= 0xf5) abort();
  if (!(*n & 0x80)) {
    v = *n;
    *_len = 1;
  } else
  if ((*n & 0xe0) == 0xc0) { /* 110xxxxx */
    if ((*(n+1) & 0xc0) != 0x80) abort();
    v = ((*n & 0x1f) << 6) | (*(n+1) & 0x3f);
    if (v < 0x80 || v > 0x7ff) abort();
    *_len = 2;
  } else
  if ((*n & 0xf0) == 0xe0) { /* 1110xxxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80) abort();
    v = ((*n & 0x0f) << (6*2)) | ((*(n+1) & 0x3f) << 6) | (*(n+2) & 0x3f);
    if (v < 0x800 || v > 0xffff) abort();
    *_len = 3;
  } else
  if ((*n & 0xf8) == 0xf0) { /* 11110xxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80 ||
        (*(n+3) & 0xc0) != 0x80) abort();
    v = ((*n & 0x07) << (6*3)) | ((*(n+1) & 0x3f) << (6*2)) |
        ((*(n+2) & 0x3f) << 6) | (*(n+3) & 0x3f);
    if (v < 0x10000 || v > 0x10ffff) abort();
    *_len = 4;
  } else abort();
  return v;
}

void expose(void)
{
  XCopyArea(d, p, w, DefaultGC(d, DefaultScreen(d)), 0,0,
            WIDTH*SCALE, HEIGHT*SCALE, 0,0);
}

void paint_pos(cairo_t *c, int symbol, int board_i, int board_j, int isfront)
{
  cairo_save(c);

  if (isfront == 1 || isfront == 3) cairo_translate(c, 1, 1);
  if (isfront == 2 || isfront == 3) cairo_translate(c, 0, 35);
  cairo_translate(c, board_j*42, board_i*42);
  cairo_rectangle(c, 0, 0, 42, 42);
  cairo_clip (c);
  cairo_set_source_surface(c, letters, -(symbol%12)*42, -symbol/12*42);
  cairo_paint(c);

  cairo_restore(c);
}

void paint(void)
{
  char t[512];
  int u2p[] = { 'a', 'b', 'c', 231, 'd', 'e', 'f', 'g', 287, 'h', 305, 'i',
    'j', 'k', 'l', 'm', 'n', 'o', 246, 'p', 'r', 's', 351, 't',
    'u', 252, 'v', 'y', 'z' };
  int i, j;
  char *g;
  int len;
  cairo_t *c;

  c=cairo_create(cs);

  cairo_scale(c, SCALE, SCALE);
  cairo_save(c);

  /* clear screen */
  cairo_rectangle(c, 0, 0, WIDTH, HEIGHT);
  cairo_set_source_rgb(c, 0.1, 0.1, 0.3);
  cairo_fill(c);

  cairo_translate(c, 5, 5+30);

  /* paint board */
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    int v = init_board[i][j];
    if (i == stary && j == starx) v = 12*9; else
    if (v == '.') v = 30; else
    if (v == 'd') v = 31; else
    if (v == 't') v = 32; else
    if (v == 'D') v = 33; else
                  v = 34;
    paint_pos(c, v, i, j, 0);
  }

  /* paint game */
  g = game;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    int s = utf8((unsigned char *)g, &len); g += len;
    int v;
    if (!s || s == '.') continue;
    for (v = 0; u2p[v] != s; v++) /**/;
    paint_pos(c, v, i, j, 1);
  }
  g = played;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    int s = utf8((unsigned char *)g, &len); g += len;
    int v;
    if (!s || s == '.') continue;
    for (v = 0; u2p[v] != s; v++) /**/;
    paint_pos(c, v+12*3, i, j, 1);
  }

  /* paint hand */
  for (i = 0; i < 7; i++) paint_pos(c, 30, 15, i+4, 2);
  g = hand;
  if (*g != '(') for (i=0; *g; i++) {
    int s = utf8((unsigned char *)g, &len); g += len;
    int v;
    for (v = 0; u2p[v] != s; v++) /**/;
    paint_pos(c, v, 15, i+4, 3);
  }

  cairo_translate(c, -5, 42*15);

  /* who's turn */
  if (meplayed) cairo_move_to(c, 15, 70);
  else          cairo_move_to(c, WIDTH-150-15, 70);
  cairo_rel_line_to(c, 150, 0);
  cairo_set_source_rgb(c, 1, 1, 0.3);
  cairo_set_line_width(c, 10);
  cairo_stroke(c);

  /* green/red rectangle to highlight the current high score */
  if (myscore > otherscore) {
    cairo_rectangle(c, 15, 10, 150, 50);
    cairo_set_source_rgb(c, .1, .7, .1);
  } else {
    cairo_rectangle(c, WIDTH-15-150, 10, 150, 50);
    cairo_set_source_rgb(c, .7, .1, .1);
  }
  cairo_fill(c);

  /* center white rectangle to print remaining letters */
  cairo_set_source_rgb(c, .9, .9, 1);
  cairo_rectangle(c, WIDTH/2-50, 3, 100, 30);
  cairo_fill(c);

  /* my score */
  sprintf(t, "%d", myscore);
  cairo_set_font_size(c, 40);
  cairo_move_to(c, 20, 50);
  cairo_show_text(c, t);

  /* other score */
  sprintf(t, "%d", otherscore);
  cairo_set_font_size(c, 40);
  cairo_move_to(c, WIDTH-150-15+5, 50);
  cairo_show_text(c, t);

  /* remain */
  sprintf(t, "%d", remain);
  cairo_set_font_size(c, 20);
  cairo_set_source_rgb(c, 0, 0, 0);
  cairo_move_to(c, WIDTH/2-20, 26);
  cairo_show_text(c, t);

  /* points */
  sprintf(t, "%d", points);
  cairo_set_font_size(c, 24);
  cairo_set_source_rgb(c, .7, .7, 1);
  if (meplayed) cairo_move_to(c, WIDTH/2-120, 28);
  else          cairo_move_to(c, WIDTH/2+80, 26);
  cairo_show_text(c, t);

  /* opponent/winner */
  cairo_restore(c);
  sprintf(t, "opponent: %s    winner: %s    %d/%d",other,winner,pos+1,npos);
  cairo_set_font_size(c, 20);
  cairo_set_source_rgb(c, .6, .6, 1);
  cairo_move_to(c, 10, 26);
  cairo_show_text(c, t);

  cairo_destroy(c);

  /* and put it all to the xpixmap */
  c=cairo_create(csout);
  cairo_set_source_surface(c, cs, 0, 0);
  cairo_rectangle(c, 0, 0, WIDTH*SCALE, HEIGHT*SCALE);
  cairo_paint(c);
  cairo_destroy(c);
}

char in[4096];
int insize;
int inpos;

void infeed(void)
{
  if (insize) abort();
  insize = read(0, in, 4096); if (insize <= 0) abort();
  inpos = 0;
}

char getin(void)
{
  if (!insize) infeed();
  insize--;
  return in[inpos++];
}

int getint(void)
{
  int ret = 0;
  int sign = 1;
  int c;
  while (1) { c = getin(); if (c != ' ') break; }
  if (c == '-') { sign = -1; c = getin(); }
  while (c >= '0' && c <= '9') { ret *= 10; ret += c - '0'; c = getin(); }
  return ret * sign;
}

void getstate(void)
{
  int i;
  for (i=0;;i++)  { int c = getin(); if (c == '\n') break; game[i] = c; }
  for (i=0;;i++)  { int c = getin(); if (c == '\n') break; played[i] = c; }
  for (i=0;;i++)  { int c = getin(); if (c == '\n') break; hand[i] = c; }
  hand[i] = 0;
  for (i=0;;i++)  { int c = getin(); if (c == '\n') break; other[i] = c; }
  for (i=0;;i++)  { int c = getin(); if (c == '\n') break; winner[i] = c; }
  myscore = getint();
  otherscore = getint();
  remain = getint();
  points = getint();
  meplayed = getint();
  starx = getint();
  stary = getint();
  pos = getint();
  npos = getint();
  while(1){ int c=getin(); if (c==':') break; }
  getin();
}

void usage(void)
{
  printf("gui [options]\n");
  printf("    run in interactive mode\n");
  printf("gui [options] <pdf file>\n");
  printf("    generate a pdf file of the game\n");
  printf("options:\n");
  printf("-scale <scale value (default 0.5)>\n");
  printf("    set the scale value\n");
  exit(0);
}

int main(int n, char **v)
{
  fd_set r;
  int redraw;
  char *pdffile = 0;
  int i;

  for (i = 1; i < n; i++) {
    if (!strcmp(v[i], "-scale"))
      { if(i>n-2)usage(); SCALE=atof(v[++i]); continue; }
    if (pdffile) usage();
    pdffile = v[i];
  }

  letters = cairo_image_surface_create_from_png("letters.png");

  cs = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,
                                  WIDTH*SCALE, HEIGHT*SCALE);

  if (pdffile) goto pdf;

  d = XOpenDisplay(0); if (!d) abort();
  w = XCreateSimpleWindow(d, DefaultRootWindow(d), 0, 0,
                          WIDTH*SCALE, HEIGHT*SCALE, 0, 0, 0);
  XSelectInput(d, w, KeyPressMask | ExposureMask);
  XMapWindow(d, w);

  p = XCreatePixmap(d, w, WIDTH*SCALE,HEIGHT*SCALE,
                    DefaultDepth(d, DefaultScreen(d)));
  csout = cairo_xlib_surface_create(d, p, DefaultVisual(d, DefaultScreen(d)),
                                    WIDTH*SCALE, HEIGHT*SCALE);

  redraw = 0;

  while (1) {
    XEvent ev;
    while (XPending(d)) {
      XNextEvent(d, &ev);
      switch (ev.type) {
      case Expose: redraw = 1; break;
      case KeyPress: {
        char buf[65];
        KeySym ks;
        XLookupString(&ev.xkey, buf, 64, &ks, NULL);
        switch (ks) {
          case XK_Right: case 'n': printf("n"); fflush(stdout); break;
          case XK_Left:  case 'p': printf("p"); fflush(stdout); break;
          case XK_Escape: case 'q': exit(1);
        }
        break;
      }
      default: break;
      }
    }
    if (insize) { while (insize) getstate(); paint(); redraw = 1; }
    if (redraw) { expose(); redraw = 0; }
    FD_ZERO(&r);
    FD_SET(0, &r);
    FD_SET(ConnectionNumber(d), &r);
    XFlush(d);
    if (select(ConnectionNumber(d)+1, &r, 0, 0, 0) == -1) abort();
    if (FD_ISSET(0, &r)) infeed();
  }

  return 0;

pdf:
  csout = cairo_pdf_surface_create(pdffile, WIDTH*SCALE, HEIGHT*SCALE);

  while (1) {
    int c = getin(); if (c == '+') break; inpos--; insize++;
    getstate();
    paint();
    cairo_surface_show_page(csout);
  }

  cairo_surface_destroy(csout);

  return 0;
}
