#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include "net.h"
#include "mess.h"
#include "kelimelik.h"

char *vraw;
FILE *vraw_file;
int verbose;

int bot;

static char *default_server = "185.15.43.107";
static int default_port = 1445;

void login(kelimelik *k)
{
  mess *m = new_mess("GameModule_requestRetrieveAccount");
  mess_arg_string(m, k->email);
  mess_arg_string(m, k->pass);
  net_send(k, m);
  free_mess(m);

  lw_lock(&k->lock);
  while (!(k->flags & USER_ID_SET)) lw_wait(&k->lock);
  lw_unlock(&k->lock);

  m = new_mess("GameModule_requestLogin");
  mess_arg_int(m, k->user_id);
  mess_arg_string(m, k->pass);
  mess_arg_int(m, 0x14);       /* I don't know what that is */
  net_send(k, m);
  free_mess(m);

  lw_lock(&k->lock);
  while (!(k->flags & LOGGED_IN)) lw_wait(&k->lock);
  lw_unlock(&k->lock);
}

void lsend(void)
{
  printf("GameModule_requestGameInfo i<game ID>\n");
  printf("GameModule_requestSubmitLetters i<gameID> s<letters> i<line> i<column> i<vertical>\n");
  printf("GameModule_requestSkipTurn i<game ID>\n");
  printf("GameModule_requestSwap i<game ID> s<letters>\n");
  exit(0);
}

void usage(void)
{
  printf("usage (each command takes [options]):\n");
  printf("kelimelik -lg\n");
  printf("    list current games and quit\n");
  printf("kelimelik -elg\n");
  printf("    list current games (and scores) and quit (slower than -lg)\n");
  printf("kelimelik -lm\n");
  printf("    list current games with messages and quit\n");
  printf("kelimelik -li\n");
  printf("    list waiting invitations and quit\n");
  printf("kelimelik -ai <invitation ID>\n");
  printf("    accept invitation and quit\n");
  printf("kelimelik -ri <invitation ID>\n");
  printf("    reject invitation and quit\n");
  printf("kelimelik -sg <game ID>\n");
  printf("    show given game and quit\n");
  printf("kelimelik -play <game ID> <letters> <line> <column> <vertical>\n");
  printf("    show given game and loop (quit if bot)\n");
  printf("kelimelik -pass <game ID>\n");
  printf("    pass given game and loop (quit if bot)\n");
  printf("kelimelik -swap <game ID> <letters>\n");
  printf("    swap letters of given game and loop (quit if bot)\n");
  printf("kelimelik -m <game ID>\n");
  printf("    show messages of given game and quit\n");
  printf("kelimelik -sm <game ID> <mess>\n");
  printf("    send message to given game and quit (loop if -v)\n");
  printf("kelimelik -iu <user name>\n");
  printf("    invite given user\n");
  printf("kelimelik -im <user email>\n");
  printf("    invite given user\n");
  printf("kelimelik -ir\n");
  printf("    invite random user\n");
  printf("kelimelik -irf\n");
  printf("    invite random user for fast game\n");
  printf("kelimelik -stats\n");
  printf("    list ended games and quit\n");
  printf("kelimelik\n");
  printf("    get a command line after login (don't lo"
         "gin if -nologin is passed)\n");
  printf("\noptions:\n");
  printf("    -u <user email address>\n");
  printf("    -p <user password>\n");
  printf("    -s <server IP> (default %s)\n", default_server);
  printf("    -sp <server port> (default %d\n", default_port);
  printf("    -nologin\n");
  printf("        don't run the login process (so -u and -p not necessary)\n");
  printf("    -l\n");
  printf("        list known messages (with syntax) you can send\n");
  printf("    -w\n");
  printf("        don't exit, wait for ever\n");
  printf("    -v\n");
  printf("        be verbose\n");
  printf("    -vraw <raw file>\n");
  printf("        dump raw received packets to file\n");
  printf("    -bot\n");
  printf("        output formatted for the bot and 30s time limit\n");
  exit(0);
}

#define PUT(c) do { \
    if (size == maxsize) { \
      buf = realloc(buf, maxsize += 512); if (!buf) abort(); \
    } \
    buf[size] = c; \
    size++; \
  } while (0)

static char *buf;
static int size;
static int maxsize;
mess *read_message(void)
{
  int c;
  mess *m;
  char *s, *t;

  size = 0;

  while (1) {
    c = getchar(); if (c == EOF) break; if (c == ';') break;
    if (isspace(c) && size && buf[size-1] == ' ') continue;
    if (isspace(c) && !size) continue;
    if (isspace(c)) c = ' ';
    PUT(c);
  }
  PUT(0);
  PUT(0);
  if (size==2) return 0;

  s = buf; t = s; while (*t && *t != ' ') t++;
  if (*t) *t = 0;

  m = new_mess(s);

  s = t+1;

  while (*s) {
    if (*s == 'i') {
      int v = 0;
      s++;
      while(*s >= '0' && *s <= '9') {
        v *= 10;
        v += *s - '0';
        s++;
      }
      if (*s && *s != ' ') goto err;
      s++;
      mess_arg_int(m, v);
      continue;
    }
    if (*s == 's') {
      char t[128];
      int i = 0;
      s++;
      while (*s && *s != ' ') { t[i] = *s; s++; i++; if (i == 127) goto err; }
      s++;
      t[i] = 0;
      mess_arg_string(m, t);
      continue;
    }
    goto err;
  }

  return m;

err:
    free_mess(m);
    printf("bad mess '%s'\n", buf);
    return 0;
}

#undef PUT

void alrm(int x)
{
  unlink("bot.log");
  unlink("botgame.log");
  exit(1);
}

int main(int n, char **v)
{
  char *server = default_server;
  int port = default_port;
  char *email = 0;
  char *pass = 0;
  int i;
  kelimelik k;
  int nologin = 0;
  int do_wait = 0;
  int list_current_games = 0;
  int list_messages = 0;
  int list_invitations = 0;
  int accept_invitation = 0;
  int reject_invitation = 0;
  int game = 0;
  int show_game = 0;
  int mess_game = 0;
  int send_mess = 0;
  char *mess_to_send;
  int invite_user = 0;
  int invite_mail = 0;
  int invite_random = 0;
  int invite_fast_random = 0;
  int play = 0;
  int turnpass = 0;
  int turnswap = 0;
  int extended_lg = 0;
  int stats = 0;
  char *user;
  char *letters;
  int line, column, vertical;

  for (i = 1; i < n; i++) {
    if (!strcmp(v[i], "-u")) { if(i>n-2)usage(); email = v[++i]; continue; }
    if (!strcmp(v[i], "-p")) { if(i>n-2)usage(); pass = v[++i]; continue; }
    if (!strcmp(v[i], "-s")) { if(i>n-2)usage(); server = v[++i]; continue; }
    if (!strcmp(v[i], "-sp")){ if(i>n-2)usage(); port=atoi(v[++i]); continue; }
    if (!strcmp(v[i], "-nologin")) { nologin = 1; continue; }
    if (!strcmp(v[i], "-w")) { do_wait = 1; continue; }
    if (!strcmp(v[i], "-v")) { verbose = 1; continue; }
    if (!strcmp(v[i], "-l")) lsend();
    if (!strcmp(v[i], "-lg")) { list_current_games = 1; continue; }
    if (!strcmp(v[i], "-lm")) { list_messages = 1; continue; }
    if (!strcmp(v[i], "-li")) { list_invitations = 1; continue; }
    if (!strcmp(v[i], "-ai"))
      { if(i>n-2)usage(); game=atoi(v[++i]); accept_invitation = 1; continue; }
    if (!strcmp(v[i], "-ri"))
      { if(i>n-2)usage(); game=atoi(v[++i]); reject_invitation = 1; continue; }
    if (!strcmp(v[i], "-sg"))
      { if(i>n-2)usage(); game=atoi(v[++i]); show_game = 1; continue; }
    if (!strcmp(v[i], "-m"))
      { if(i>n-2)usage(); game=atoi(v[++i]); mess_game = 1; continue; }
    if (!strcmp(v[i], "-sm"))
      { if(i>n-3)usage(); game=atoi(v[++i]); mess_to_send=v[++i];
        send_mess = 1; continue; }
    if (!strcmp(v[i], "-vraw")){ if(i>n-2)usage(); vraw=v[++i]; continue; }
    if (!strcmp(v[i], "-iu"))
      { if(i>n-2)usage(); user=v[++i]; invite_user = 1; continue; }
    if (!strcmp(v[i], "-im"))
      { if(i>n-2)usage(); user=v[++i]; invite_mail = 1; continue; }
    if (!strcmp(v[i], "-ir")) { invite_random = 1; continue; }
    if (!strcmp(v[i], "-irf")) { invite_fast_random = 1; continue; }
    if (!strcmp(v[i], "-play")) { if(i>n-5)usage(); game=atoi(v[++i]);
      letters=v[++i]; line=atoi(v[++i]); column=atoi(v[++i]);
      vertical=atoi(v[++i]); play = 1; continue; }
    if (!strcmp(v[i], "-pass"))
      { if(i>n-2)usage(); game=atoi(v[++i]); turnpass = 1; continue; }
    if (!strcmp(v[i], "-swap"))
      {if(i>n-3)usage();game=atoi(v[++i]);letters=v[++i];turnswap=1;continue;}
    if (!strcmp(v[i], "-bot")) { bot = 1; continue; }
    if (!strcmp(v[i], "-elg"))
      { list_current_games = 1; extended_lg = 1; continue; }
    if (!strcmp(v[i], "-stats")) { stats = 1; continue; }
    usage();
  }
  if (vraw){vraw_file=fopen(vraw,"w");if(!vraw_file){perror("vraw");return 1;}}
  if (!nologin) if (!email || !pass) usage();
  if (list_current_games && nologin) usage();
  if (list_messages && nologin) usage();
  if (list_invitations && nologin) usage();
  if (accept_invitation && nologin) usage();
  if (reject_invitation && nologin) usage();
  if (show_game && nologin) usage();
  if (mess_game && nologin) usage();
  if (invite_user && nologin) usage();
  if (invite_mail && nologin) usage();
  if (invite_random && nologin) usage();
  if (invite_fast_random && nologin) usage();
  if (play && nologin) usage();
  if (pass && nologin) usage();
  if (stats && nologin) usage();

  if (bot) {
    signal(SIGALRM, alrm);
    alarm(30);
  }

  k.email  = email;
  k.pass   = pass;
  k.server = server;
  k.port   = port;
  k.flags  = 0;
  lw_init(&k.lock);

  net_open(&k);
  if (!nologin) login(&k);

  if (turnswap) {
    mess *m = new_mess("GameModule_requestSwap");
    if (!bot) verbose = 1;
    mess_arg_int(m, game);
    mess_arg_string(m, letters);
    net_send(&k, m);
    free_mess(m);
    if (bot) { sleep(2); return 0; }
    goto loop;
  }

  if (turnpass) {
    mess *m = new_mess("GameModule_requestSkipTurn");
    if (!bot) verbose = 1;
    mess_arg_int(m, game);
    net_send(&k, m);
    free_mess(m);
    if (bot) { sleep(2); return 0; }
    goto loop;
  }

  if (play) {
    mess *m = new_mess("GameModule_requestSubmitLetters");
    if (!bot) verbose = 1;
    mess_arg_int(m, game);
    mess_arg_string(m, letters);
    mess_arg_int(m, line);
    mess_arg_int(m, column);
    mess_arg_int(m, vertical);
    net_send(&k, m);
    free_mess(m);
    if (bot) {
      lw_lock(&k.lock);
      while (!((k.flags & WORD_OK) || (k.flags & WORD_NOOK))) lw_wait(&k.lock);
      lw_unlock(&k.lock);
      printf("%d\n", !!(k.flags & WORD_OK));
      return 0;
    }
    goto loop;
  }

  if (accept_invitation) {
    mess *m = new_mess("GameModule_requestAcceptInvitation");
    mess_arg_int(m, game);
    net_send(&k, m);
    free_mess(m);
    if (verbose) goto loop;
    return 0;
  }

  if (reject_invitation) {
    mess *m = new_mess("GameModule_requestRejectInvitation");
    mess_arg_int(m, game);
    net_send(&k, m);
    free_mess(m);
    if (verbose) goto loop;
    return 0;
  }

  if (invite_user) {
    mess *m = new_mess("GameModule_requestGameInvitationToUser");
    mess_arg_string(m, user);
    net_send(&k, m);
    free_mess(m);
    if (verbose) goto loop;
    return 0;
  }

  if (invite_mail) {
    mess *m = new_mess("GameModule_requestGameInvitationToEmail");
    mess_arg_string(m, user);
    net_send(&k, m);
    free_mess(m);
    if (verbose) goto loop;
    return 0;
  }

  if (invite_random) {
    mess *m = new_mess("GameModule_requestCreateGameRandom");
    net_send(&k, m);
    free_mess(m);
    return 0;
  }

  if (invite_fast_random) {
    mess *m = new_mess("GameModule_requestCreateFastGameRandom");
    net_send(&k, m);
    free_mess(m);
    return 0;
  }

  if (send_mess) {
    mess *m = new_mess("GameModule_requestSendMessage");
    mess_arg_int(m, game);
    mess_arg_string(m, mess_to_send);
    net_send(&k, m);
    free_mess(m);
    if (verbose) goto loop;
    return 0;
  }

  if (mess_game) {
    int n, i;
    mess *m = new_mess("GameModule_requestGameMessages");
    mess_arg_int(m, game);
    net_send(&k, m);
    free_mess(m);

    lw_lock(&k.lock);
    while (!(k.flags & MESSAGES_OK)) lw_wait(&k.lock);
    lw_unlock(&k.lock);

    if (!check_mess(k.messages, 4, INT,
        ARRAY, INT, 1, ARRAY, STRING, 1, ARRAY, DATE, 1))
      goto mg_err;

    n = k.messages->args[1].a->size;
    for (i = 0; i < n; i++) {
      time_t t;
      char tb[26];
      t = k.messages->args[3].a->l[i].d;
      ctime_r(&t, tb); tb[24] = 0;
      printf("[%s] %d: %s\n", tb, k.messages->args[1].a->l[i].i,
             k.messages->args[2].a->l[i].s);
    }

    return 0;
mg_err:
    printf("bad message mess\n"); print_mess(k.messages); return 0;
  }

  if (show_game) {
    int myturn;
    int starx, stary;
    int new[15*15];
    int i, j;
    char *s;
    mess *m = new_mess("GameModule_requestGameInfo");
    mess_arg_int(m, game);
    net_send(&k, m);
    free_mess(m);

    for(i=0;i<15*15;i++)new[i]=0;

    lw_lock(&k.lock);
    while (!(k.flags & GAME_OK)) lw_wait(&k.lock);
    lw_unlock(&k.lock);

    if (!check_mess(k.game, 12, INT, INT, INT, INT, INT, STRING, INT,
        STRING, STRING, INT, INT, INT)
        && !check_mess(k.game, 12, INT, INT, INT, INT, INT, STRING, INT,
            STRING, STRING, INT, INT, UNKNOWN_1))
      goto sg_err;

    myturn = k.game->args[2].i;
    starx = (k.game->args[10].i-1) / 15;
    stary = (k.game->args[10].i-1) % 15;

    s = k.game->args[5].s;
    if (s) while (*s) {
      int z;
      while (*s && !isdigit(*s)) s++;
      if (sscanf(s, "%d", &z) == 1 && z <= 15*15) new[z] = 1;
      while (*s && isdigit(*s)) s++;
    }

    s = k.game->args[8].s;
    for (i = 0; i < 15; i++) {
      for (j = 0; j < 15; j++) {
        int l;
        if (!bot && new[i*15+j]) printf("\033[42;30m");
        if (*s == 0) goto sg_err;
        if ((*s) & 0x80) {
          if ((*s & 0xe0) == 0xc0) l = 2; else
          if ((*s & 0xf0) == 0xe0) l = 3; else
          if ((*s & 0xf8) == 0xf0) l = 4; else goto sg_err;
          for (; *s && l; l--, s++) printf("%c", *s);
        } else {
          if (!bot && *s == '.' && i == starx && j == stary)
            printf("*");
          else
            printf("%c", *s);
          s++;
        }
        if (!bot && new[i*15+j]) printf("\033[m");
      }
      printf("\n");
    }
    if (bot) {
      char *oname;
      char *winner;
      int n;
      int op;
      char *score[15] = {
        "..T..d...d..T..",
        ".t....d.d....t.",
        "T......D......T",
        "...D.......D...",
        "....t.....t....",
        "d....d...d....d",
        ".d....d.d....d.",
        "..D.........D..",
        ".d....d.d....d.",
        "d....d...d....d",
        "....t.....t....",
        "...D.......D...",
        "T......D......T",
        ".t....d.d....t.",
        "..T..d...d..T.."
      };

      lw_lock(&k.lock);
      while (!(k.flags & COMPLETED_GAME_LIST_OK)) lw_wait(&k.lock);
      while (!(k.flags & GAME_LIST_OK)) lw_wait(&k.lock);
      lw_unlock(&k.lock);

      if (!check_mess(k.game_list_mess, 11, ARRAY, INT, 0, ARRAY, STRING, 0,
          ARRAY, INT, 0, ARRAY, STRING, 0, ARRAY, STRING, 0,
          ARRAY, INT, 0, ARRAY, DATE, 0, ARRAY, INT, 0, ARRAY, INT, 0,
          ARRAY, UNKNOWN_1, 0, ARRAY, UNKNOWN_1, 0)) abort();
      if (!check_mess(k.completed_game_list_mess, 13, ARRAY, INT, 0,
          ARRAY, STRING, 0, ARRAY, INT, 0, ARRAY, STRING, 0,
          ARRAY, INT, 0, ARRAY, STRING, 0, ARRAY, INT, 0, ARRAY, INT, 0,
          ARRAY, DATE, 0, ARRAY, INT, 0, INT, INT,
          ARRAY, UNKNOWN_1, 0)) abort();

      oname = 0;
      winner = 0;
      op = k.game->args[1].i;
      n = k.game_list_mess->args[0].a->size;
      for (i = 0; i < n; i++) {
        array *opname = k.game_list_mess->args[1].a;
        array *opid= k.game_list_mess->args[2].a;
        if (opid->l[i].i == op) oname = opname->l[i].s;
      }
      n = k.completed_game_list_mess->args[0].a->size;
      for (i = 0; i < n; i++) {
        array *game_id = k.completed_game_list_mess->args[0].a;
        array *opname = k.completed_game_list_mess->args[1].a;
        array *opid= k.completed_game_list_mess->args[2].a;
        array *win = k.completed_game_list_mess->args[3].a;
        if (opid->l[i].i == op) oname = opname->l[i].s;
        if (game_id->l[i].i == game) winner = win->l[i].s;
      }
      if (!oname) oname = "";
      if (!winner) winner = "";

      for (i = 0; i < 15; i++) {
        for (j = 0; j < 15; j++) {
          if (i == starx && j == stary) printf("s");
          else printf("%c", score[i][j]);
        }
        printf("\n");
      }
      printf("%s.......\n", k.game->args[7].s);
      /* star pos x, star pos y, your score, opponent score, remaining letters,
       * opponent string, my letters, opname length, opname,
       * winner length, winner */
      printf("%d %d %d %d %d [%s] [%s] %d %s %d %s\n", starx, stary,
             k.game->args[3].i, k.game->args[4].i, k.game->args[6].i,
             k.game->args[5].s,
             k.game->args[7].s, (int)strlen(oname), oname,
             (int)strlen(winner), winner);
      return 0;
    }
    printf("%s %d %s %d [star %d.%d] letters %s (%d remain)\n",
           myturn ? "*you*" : "you",
           k.game->args[3].i,
           myturn ? "opponent" : "*opponent*",
           k.game->args[4].i,
           starx, stary,
           k.game->args[7].s,
           k.game->args[6].i);

    return 0;
sg_err:
    printf("bad game mess\n"); print_mess(k.game); return 0;
  }

  if (list_current_games) {
    int n;
    lw_lock(&k.lock);
    while (!(k.flags & GAME_LIST_OK)) lw_wait(&k.lock);
    lw_unlock(&k.lock);

    if (!check_mess(k.game_list_mess, 11, ARRAY, INT, 0, ARRAY, STRING, 0,
        ARRAY, INT, 0, ARRAY, STRING, 0, ARRAY, STRING, 0,
        ARRAY, INT, 0, ARRAY, DATE, 0, ARRAY, INT, 0, ARRAY, INT, 0,
        ARRAY, UNKNOWN_1, 0, ARRAY, UNKNOWN_1, 0))
      goto lcg_err;

    n = k.game_list_mess->args[0].a->size;

    for (i = 0; i < n; i++) {
      int star_free;
      char tb[26];
      char your_score[128];
      char opponent_score[128];
      int remain;
      time_t t;
      int ii, j;
      array *game_id = k.game_list_mess->args[0].a;
      array *opponent_name = k.game_list_mess->args[1].a;
      array *last_word = k.game_list_mess->args[4].a;
      array *last_score = k.game_list_mess->args[5].a;
      array *last_date = k.game_list_mess->args[6].a;
      array *your_turn = k.game_list_mess->args[7].a;
      array *new_message = k.game_list_mess->args[8].a;
      /* barbarian sort */
      ii = i;
      i = -1;
      if (bot) {
        for (j = 0; j < n; j++) {
          if (last_date->l[j].d == 0) continue;
          if (i == -1 || last_date->l[j].d < last_date->l[i].d) i = j;
        }
      } else {
        for (j = 0; j < n; j++) {
          if (last_date->l[j].d == 0) continue;
          if (i == -1 || last_date->l[j].d > last_date->l[i].d) i = j;
        }
      }
      t = last_date->l[i].d;
      ctime_r(&t, tb); tb[24] = 0;
      if (!bot && extended_lg) {
        char *s;
        int starx, stary, ii, jj;
        int other, me;
        mess *m = new_mess("GameModule_requestGameInfo");
        mess_arg_int(m, game_id->l[i].i);

        lw_lock(&k.lock);
        k.flags &= ~GAME_OK;
        lw_unlock(&k.lock);

        net_send(&k, m);
        free_mess(m);

        lw_lock(&k.lock);
        while (!(k.flags & GAME_OK)) lw_wait(&k.lock);
        lw_unlock(&k.lock);

        if (!check_mess(k.game, 12, INT, INT, INT, INT, INT, STRING, INT,
            STRING, STRING, INT, INT, INT)
            && !check_mess(k.game, 12, INT, INT, INT, INT, INT, STRING, INT,
                STRING, STRING, INT, INT, UNKNOWN_1)) abort();

        starx = (k.game->args[10].i-1) / 15;
        stary = (k.game->args[10].i-1) % 15;

        /* check if the star is free (barbarian) */
        star_free = 0;
        s = k.game->args[8].s;
        for (ii = 0; ii < 15; ii++) {
          for (jj = 0; jj < 15; jj++) {
            int l;
            if (*s == 0) abort();
            if ((*s) & 0x80) {
              if ((*s & 0xe0) == 0xc0) l = 2; else
              if ((*s & 0xf0) == 0xe0) l = 3; else
              if ((*s & 0xf8) == 0xf0) l = 4; else abort();
              for (; *s && l; l--, s++) /**/;
            } else {
              if (*s == '.' && ii == starx && jj == stary)
                star_free++;
              s++;
            }
          }
        }
        me = k.game->args[3].i;
        other = k.game->args[4].i;
        remain = k.game->args[6].i;

        if (me > other) {
          sprintf(your_score, "\033[42;30m%3d\033[m", me);
          sprintf(opponent_score, "%3d", other);
        } else {
          sprintf(your_score, "%3d", me);
          sprintf(opponent_score, "\033[41;30m%3d\033[m", other);
        }

        free_mess(k.game);
      }
      if (bot) printf("%d %d\n", game_id->l[i].i, your_turn->l[i].i);
      else {
        if (extended_lg)
          printf("%d %s%s%c%2d (%3d) %s [%s] [%s] %s %s\n",
                 game_id->l[i].i, your_score, opponent_score,
                 star_free ? '*' : ' ', remain,
                 last_score->l[i].i,
                 opponent_name->l[i].s,
                 last_word->l[i].s,
                 tb,
                 your_turn->l[i].i ? "\033[41;30m[YOU]\033[m" : "",
                 new_message->l[i].i ? "\033[42;30m[MESS]\033[m" : "");
        else
          printf("%d (%3d) %s [%s] [%s] %s %s\n",
                 game_id->l[i].i,
                 last_score->l[i].i,
                 opponent_name->l[i].s,
                 last_word->l[i].s,
                 tb,
                 your_turn->l[i].i ? "\033[41;30m[YOU]\033[m" : "",
                 new_message->l[i].i ? "\033[42;30m[MESS]\033[m" : "");
      }
      last_date->l[i].d = 0;
      i = ii;
    }
    return 0;
lcg_err:
    printf("bad game list mess\n"); print_mess(k.game_list_mess); return 0;
  }

  if (list_messages) {
    int n;
    lw_lock(&k.lock);
    while (!((k.flags & GAME_LIST_OK) && (k.flags & COMPLETED_GAME_LIST_OK)))
      lw_wait(&k.lock);
    lw_unlock(&k.lock);

    if (!check_mess(k.game_list_mess, 11, ARRAY, INT, 0, ARRAY, STRING, 0,
        ARRAY, INT, 0, ARRAY, STRING, 0, ARRAY, STRING, 0,
        ARRAY, INT, 0, ARRAY, DATE, 0, ARRAY, INT, 0, ARRAY, INT, 0,
        ARRAY, UNKNOWN_1, 0, ARRAY, UNKNOWN_1, 0))
      goto lm_err;

    if (!check_mess(k.completed_game_list_mess, 13, ARRAY, INT, 0,
        ARRAY, STRING, 0, ARRAY, INT, 0, ARRAY, STRING, 0,
        ARRAY, INT, 0, ARRAY, STRING, 0, ARRAY, INT, 0, ARRAY, INT, 0,
        ARRAY, DATE, 0, ARRAY, INT, 0, INT, INT,
        ARRAY, UNKNOWN_1, 0))
      goto lm_err;

    n = k.game_list_mess->args[0].a->size;
    for (i = 0; i < n; i++) {
      array *game_id = k.game_list_mess->args[0].a;
      array *opponent_name = k.game_list_mess->args[1].a;
      array *new_message = k.game_list_mess->args[8].a;
      if (new_message->l[i].i)
        printf("%d %s\n", game_id->l[i].i, opponent_name->l[i].s);
    }

    n = k.completed_game_list_mess->args[0].a->size;
    for (i = 0; i < n; i++) {
      array *game_id = k.completed_game_list_mess->args[0].a;
      array *opponent_name = k.completed_game_list_mess->args[1].a;
      array *new_message = k.completed_game_list_mess->args[9].a;
      if (new_message->l[i].i)
        printf("%d %s\n", game_id->l[i].i, opponent_name->l[i].s);
    }

    return 0;

lm_err:
    printf("error\n"); return 0;
  }

  if (stats) {
    int n;
    lw_lock(&k.lock);
    while (!((k.flags & COMPLETED_GAME_LIST_OK) && (k.flags&USER_PROFILE_OK)))
      lw_wait(&k.lock);
    lw_unlock(&k.lock);

    if (!check_mess(k.completed_game_list_mess, 13, ARRAY, INT, 0,
        ARRAY, STRING, 0, ARRAY, INT, 0, ARRAY, STRING, 0,
        ARRAY, INT, 0, ARRAY, STRING, 0, ARRAY, INT, 0, ARRAY, INT, 0,
        ARRAY, DATE, 0, ARRAY, INT, 0, INT, INT,
        ARRAY, UNKNOWN_1, 0))
      goto stats_err;

    if (!check_mess(k.user, 3, INT, INT, INT)) goto stats_err;

    n = k.completed_game_list_mess->args[0].a->size;

    if (bot) {
      for (i = 0; i < n; i++) {
        array *game_id = k.completed_game_list_mess->args[0].a;
        printf("%d\n", game_id->l[i].i);
      }
      return 0;
    }

    for (i = 0; i < n; i++) {
      char score[256];
      char tb[26];
      time_t t;
      int ii, j;
      array *game_id = k.completed_game_list_mess->args[0].a;
      array *opponent_name = k.completed_game_list_mess->args[1].a;
      array *new_message = k.completed_game_list_mess->args[9].a;
      array *you_score = k.completed_game_list_mess->args[6].a;
      array *opponent_score = k.completed_game_list_mess->args[7].a;
      array *date = k.completed_game_list_mess->args[8].a;
      array *winner = k.completed_game_list_mess->args[3].a;
      ii = -1;
      /* barbarian sort */
      for (j = 0; j < n; j++) {
        if (date->l[j].d == 0) continue;
        if (ii == -1 || date->l[j].d > date->l[ii].d) ii = j;
      }
      if (you_score->l[ii].i > opponent_score->l[ii].i)
        sprintf(score, "\033[42;30m%3d\033[m%3d", you_score->l[ii].i,
                opponent_score->l[ii].i);
      else
        sprintf(score, "%3d\033[41;30m%3d\033[m", you_score->l[ii].i,
                opponent_score->l[ii].i);
      t = date->l[ii].d;
      ctime_r(&t, tb); tb[24] = 0;
      printf("%s%d\033[m %s %s %s [%s]\n",
          new_message->l[ii].i?"\033[42;30m":"", game_id->l[ii].i,
          score, tb, opponent_name->l[ii].s, winner->l[ii].s);
      date->l[ii].d = 0;
    }
    printf("%d games played, %d games won\n",
           k.user->args[1].i, k.user->args[2].i);

    return 0;

stats_err:
    printf("error\n"); return 0;
  }

  if (list_invitations) {
    int n;
    lw_lock(&k.lock);
    while (!(k.flags & INVITATIONS_OK))
      lw_wait(&k.lock);
    lw_unlock(&k.lock);

    if (!check_mess(k.invit, 4, ARRAY, INT, 0, ARRAY, STRING, 0,
        ARRAY, STRING, 0, ARRAY, DATE, 0))
      goto li_err;

    n = k.invit->args[0].a->size;
    for (i = 0; i < n; i++) {
      array *game_id = k.invit->args[0].a;
      array *opponent_name = k.invit->args[1].a;
      printf("%d %s\n", game_id->l[i].i, opponent_name->l[i].s);
    }
    return 0;

li_err:
    printf("error\n"); return 0;
  }

loop:
  while (1) {
    mess *m;
    printf("> "); fflush(stdout);
    m = read_message(); if (!m) break;
    if (verbose) { printf("sending:\n"); print_mess(m); }
    net_send(&k, m);
    free_mess(m);
  }

  if (do_wait) while (1) pause();

  return 0;
}
