#include "mess.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>

char *arg_type_s(arg_type t)
{
  switch (t) {
  case STRING: return "STRING";
  case INT: return "INT";
  case DATE: return "DATE";
  case ARRAY: return "ARRAY";
  case UNKNOWN_1: return "UNKNOWN_1";
  case UNKNOWN_3: return "UNKNOWN_3";
  }
  abort();
}

array *new_array(arg_type t)
{
  array *a = malloc(sizeof(array)); if (!a) abort();
  a->type = t;
  a->l = 0;
  a->size = 0;
  return a;
}

void array_new_arg(array *a)
{
  a->size++;
  a->l = realloc(a->l, a->size * sizeof(arg)); if(!a->l)abort();
}

void array_arg_date(array *a, uint64_t d)
{
  array_new_arg(a);
  a->l[a->size-1].type = DATE;
  a->l[a->size-1].d = d;
}

void array_arg_int(array *a, int t)
{
  array_new_arg(a);
  a->l[a->size-1].type = INT;
  a->l[a->size-1].i = t;
}

void array_arg_string(array *a, char *s)
{
  array_new_arg(a);
  a->l[a->size-1].type = STRING;
  if (!s) { a->l[a->size-1].s = 0; return; }
  a->l[a->size-1].s = strdup(s); if (!a->l[a->size-1].s) abort();
}

mess *new_mess(char *s)
{
  mess *r = malloc(sizeof(mess)); if (!r) abort();
  r->fun = strdup(s); if (!r->fun) abort();
  r->args = 0;
  r->args_size = 0;
  return r;
}

void mess_new_arg(mess *m)
{
  m->args_size++;
  m->args = realloc(m->args, m->args_size * sizeof(arg)); if(!m->args)abort();
}

void mess_arg_string(mess *m, char *s)
{
  mess_new_arg(m);
  m->args[m->args_size-1].type = STRING;
  if (!s) { m->args[m->args_size-1].s = 0; return; }
  m->args[m->args_size-1].s = strdup(s); if(!m->args[m->args_size-1].s)abort();
}

void mess_arg_int(mess *m, int i)
{
  mess_new_arg(m);
  m->args[m->args_size-1].type = INT;
  m->args[m->args_size-1].i = i;
}

void free_mess(mess *m)
{
  int i;
  free(m->fun);
  for (i = 0; i < m->args_size; i++) {
    switch (m->args[i].type) {
    case STRING: free(m->args[i].s); break;
    case ARRAY: {
      array *a = m->args[i].a;
      int j;
      for (j = 0; j < a->size; j++) switch (a->type) {
      default: break;
      case STRING: free(a->l[j].s); break;
      }
      free(a);
      break;
    }
    default: break;
    }
  }
  free(m->args);
  free(m);
}

char *put_fun(char *b, char *s)
{
  int l = strlen(s);
  int i;
  b[0] = (l >> 8) & 255;
  b[1] = l & 255;
  b += 2;
  for (i = 0; i < l; b++, i++) *b = s[i];
  return b;
}

char *put_string(char *b, char *s)
{
  int l = strlen(s);
  int i;
  b[0] = 7;
  b[1] = (l >> 8) & 255;
  b[2] = l & 255;
  b += 3;
  for (i = 0; i < l; b++, i++) *b = s[i];
  return b;
}

char *put_int(char *b, int i)
{
  b[0] = 0;
  b[1] = (i >> 24) & 255;
  b[2] = (i >> 16) & 255;
  b[3] = (i >> 8) & 255;
  b[4] = i & 255;
  return b+5;
}

char *mess_serialize(mess *m, int *_len)
{
  char *b, *bs;
  int len;
  int i;

  /* length of packet (not including the 4 starting bytes) */
  len = 4;

  /* function length + function */
  len += 2 + strlen(m->fun);

  /* number of args */
  len++;

  /* args */
  for (i = 0; i < m->args_size; i++) {
    /* type */
    len++;
    /* data */
    switch (m->args[i].type) {
    case STRING: len += 2 + strlen(m->args[i].s); break;
    case INT: len += 4; break;
    default: abort();
    }
  }

  b = malloc(len); if (!b) abort();
  bs = b;

  b[0] = ((len-4) >> (8*3)) & 255;
  b[1] = ((len-4) >> (8*2)) & 255;
  b[2] = ((len-4) >> (8*1)) & 255;
  b[3] = ((len-4) >> (8*0)) & 255;

  b += 4;
  b = put_fun(b, m->fun);

  *b = m->args_size; b++;

  for (i = 0; i < m->args_size; i++) {
    switch (m->args[i].type) {
    case STRING: b = put_string(b, m->args[i].s); break;
    case INT: b = put_int(b, m->args[i].i); break;
    default: abort();
    }
  }

  *_len = len;
  return bs;
}

unsigned char get_byte(unsigned char **_b, int *_len)
{
  unsigned char *b = *_b;
  int len = *_len;
  unsigned char c;
  if (!len) { printf("bad in packet\n"); exit(1); }
  c = *b;
  b++;
  len--;
  *_len = len;
  *_b = b;
  return c;
}

char *get_string(unsigned char **_b, int *_len)
{
  int i;
  int l;
  char *r;
  unsigned char *b = *_b;
  int len = *_len;

  l = get_byte(&b, &len);
  l = (l << 8) | get_byte(&b, &len);
  if (l) {
    r = malloc(l+1); if (!r) abort();
    for (i = 0;  i < l; i++) r[i] = get_byte(&b, &len);
    r[l] = 0;
  }
  else r = 0;

  *_len = len;
  *_b = b;

  return r;
}

array *get_array(unsigned char **_b, int *_len)
{
  array *a;
  unsigned char *b = *_b;
  int len = *_len;
  int array_size;
  int t;
  arg_type type;
  int i;
  char *s;
  uint64_t d;

  array_size = get_byte(&b, &len);
  array_size = (array_size << 8) | get_byte(&b, &len);
  array_size = (array_size << 8) | get_byte(&b, &len);
  array_size = (array_size << 8) | get_byte(&b, &len);

  switch ((t = get_byte(&b, &len))) {
  case 0: type = INT; break;
  case 7: type = STRING; break;
  case 1: type = UNKNOWN_1; break;
  case 3: type = DATE; break;
  default: printf("unhandled array type %d\n", t); exit(1);
  }

  a = new_array(type);

  for (i = 0; i < array_size; i++) switch (type) {
  case INT:
    t = get_byte(&b, &len);
    t = (t << 8) | get_byte(&b, &len);
    t = (t << 8) | get_byte(&b, &len);
    t = (t << 8) | get_byte(&b, &len);
    array_arg_int(a, t);
    break;
  case STRING:
    s = get_string(&b, &len);
    array_arg_string(a, s);
    free(s);
    break;
  case UNKNOWN_1:
    array_arg_int(a, get_byte(&b, &len));
    break;
  case DATE:
    d = get_byte(&b, &len);
    d = (d << 8) | get_byte(&b, &len);
    d = (d << 8) | get_byte(&b, &len);
    d = (d << 8) | get_byte(&b, &len);
    d = (d << 8) | get_byte(&b, &len);
    d = (d << 8) | get_byte(&b, &len);
    d = (d << 8) | get_byte(&b, &len);
    d = (d << 8) | get_byte(&b, &len);
    array_arg_date(a, d);
    break;
  default: abort();
  }

  *_len = len;
  *_b = b;

  return a;
}

mess *mess_unserialize(char *_b, int len)
{
  unsigned char *b = (unsigned char *)_b;
  mess *m = malloc(sizeof(mess)); if (!m) abort();
  int nargs;
  int i;
  int type;

  m->fun = get_string(&b, &len);

  nargs = get_byte(&b, &len);

  m->args = malloc(sizeof(arg) * nargs); if (!m->args) abort();
  m->args_size = nargs;

  for (i = 0; i < nargs; i++) {
  switch ((type=get_byte(&b, &len))) {
  case 0:
    m->args[i].type = INT;
    m->args[i].i = get_byte(&b, &len);
    m->args[i].i = (m->args[i].i << 8) | get_byte(&b, &len);
    m->args[i].i = (m->args[i].i << 8) | get_byte(&b, &len);
    m->args[i].i = (m->args[i].i << 8) | get_byte(&b, &len);
    break;
  case 7:
    m->args[i].type = STRING;
    m->args[i].s = get_string(&b, &len);
    break;
  case 8:
    m->args[i].type = ARRAY;
    m->args[i].a = get_array(&b, &len);
    break;
  case 3:
    m->args[i].type = DATE;
    m->args[i].d = get_byte(&b, &len);
    m->args[i].d = (m->args[i].d << 8) | get_byte(&b, &len);
    m->args[i].d = (m->args[i].d << 8) | get_byte(&b, &len);
    m->args[i].d = (m->args[i].d << 8) | get_byte(&b, &len);
    m->args[i].d = (m->args[i].d << 8) | get_byte(&b, &len);
    m->args[i].d = (m->args[i].d << 8) | get_byte(&b, &len);
    m->args[i].d = (m->args[i].d << 8) | get_byte(&b, &len);
    m->args[i].d = (m->args[i].d << 8) | get_byte(&b, &len);
    break;
  case 1:
    m->args[i].type = UNKNOWN_1;
    m->args[i].i = get_byte(&b, &len);
    break;
  default: printf("bad arg type %d\n", type); exit(1);
  }
  }

  if (len) { printf("bad packet, bytes at the end\n"); exit(1); }
  return m;
}

void print_mess(mess *m)
{
  int i;
  printf("%s (%d args)\n", m->fun, m->args_size);
  for (i = 0; i < m->args_size; i++) {
  switch (m->args[i].type) {
  case INT: printf("  [INT] %d\n", m->args[i].i); break;
  case STRING: printf("  [STRING] %s\n", m->args[i].s); break;
  case ARRAY: {
    char tb[26];
    array *a = m->args[i].a;
    int j;
    printf("  array (size %d) (type %s)\n", a->size, arg_type_s(a->type));
    for (j = 0; j < a->size; j++) switch(a->type) {
    case INT: printf("    [INT] %d\n", a->l[j].i); break;
    case STRING: printf("    [STRING] %s\n", a->l[j].s); break;
    case UNKNOWN_1: printf("    [UNKNOWN_1] %d\n", a->l[j].i); break;
    case DATE: {time_t t=a->l[j].d;
                printf("    [DATE] %s", ctime_r(&t, tb)); break;}
    default: abort();
    }
    break;
  }
  case DATE: {char tb[26];
    time_t t=m->args[i].d; printf("  [DATE] %s", ctime_r(&t, tb)); break;}
  case UNKNOWN_1: printf("  [UNKNOWN_1] %d\n", m->args[i].i); break;
  default: abort();
  }
  }
}

int check_mess(mess *m, int nargs, ...)
{
  int i;
  int ret = 1;
  arg_type type, atype;
  int len_index;
  va_list l;

  if (m->args_size != nargs) return 0;

  va_start(l, nargs);

  for (i = 0; i < nargs; i++) {
    type = va_arg(l, arg_type);
    if (m->args[i].type != type) { ret = 0; break; }
    if (type != ARRAY) continue;
    atype = va_arg(l, arg_type); len_index = va_arg(l, int);
    if (m->args[i].a->type != atype ||
        m->args[i].a->size != m->args[len_index].a->size) { ret = 0; break; }
  }

  va_end(l);

  return ret;
}
