#include "net.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>

static int sock;

static unsigned char get(void)
{
  extern FILE *vraw_file;
  char c;
  int r = read(sock, &c, 1);
  if (r == -1) { perror("socket read"); exit(1); }
  if (r != 1) { printf("socket read bad return %d\n", r); exit(1); }
  if (vraw) { fputc(c, vraw_file); fflush(vraw_file); }
  return c;
}

void handle_mess(kelimelik *k, mess *m)
{
  if (!strcmp(m->fun, "GameModule_createAccountResponse")) {
    if (m->args_size != 6 ||
        m->args[0].type != INT || m->args[0].i != 1 ||
        m->args[1].type != INT) {
      print_mess(m);
      exit(1);
    }
    lw_lock(&k->lock);
    k->user_id = m->args[1].i;
    k->flags |= USER_ID_SET;
    lw_wakeup(&k->lock);
    lw_unlock(&k->lock);
    return;
  }
  if (!strcmp(m->fun, "GameModule_loginAccepted")) {
    lw_lock(&k->lock);
    k->flags |= LOGGED_IN;
    lw_wakeup(&k->lock);
    lw_unlock(&k->lock);
    return;
  }
  if (!strcmp(m->fun, "GameModule_userGamesList")) {
    lw_lock(&k->lock);
    k->game_list_mess = new_mess("GameModule_userGamesList");
    k->game_list_mess->args = m->args;
    k->game_list_mess->args_size = m->args_size;
    m->args = 0;
    m->args_size = 0;
    k->flags |= GAME_LIST_OK;
    lw_wakeup(&k->lock);
    lw_unlock(&k->lock);
    return;
  }
  if (!strcmp(m->fun, "GameModule_userCompletedGamesList")) {
    lw_lock(&k->lock);
    k->completed_game_list_mess=new_mess("GameModule_userCompletedGamesList");
    k->completed_game_list_mess->args = m->args;
    k->completed_game_list_mess->args_size = m->args_size;
    m->args = 0;
    m->args_size = 0;
    k->flags |= COMPLETED_GAME_LIST_OK;
    lw_wakeup(&k->lock);
    lw_unlock(&k->lock);
    return;
  }
  if (!strcmp(m->fun, "GameModule_userProfile")) {
    lw_lock(&k->lock);
    k->user=new_mess("GameModule_userProfile");
    k->user->args = m->args;
    k->user->args_size = m->args_size;
    m->args = 0;
    m->args_size = 0;
    k->flags |= USER_PROFILE_OK;
    lw_wakeup(&k->lock);
    lw_unlock(&k->lock);
    return;
  }
  if (!strcmp(m->fun, "GameModule_gameInfo")) {
    lw_lock(&k->lock);
    k->game = new_mess("GameModule_gameInfo");
    k->game->args = m->args;
    k->game->args_size = m->args_size;
    m->args = 0;
    m->args_size = 0;
    k->flags |= GAME_OK;
    lw_wakeup(&k->lock);
    lw_unlock(&k->lock);
    return;
  }
  if (!strcmp(m->fun, "GameModule_gameMessages")) {
    lw_lock(&k->lock);
    k->messages = new_mess("GameModule_gameMessages");
    k->messages->args = m->args;
    k->messages->args_size = m->args_size;
    m->args = 0;
    m->args_size = 0;
    k->flags |= MESSAGES_OK;
    lw_wakeup(&k->lock);
    lw_unlock(&k->lock);
    return;
  }
  if (!strcmp(m->fun, "GameModule_wordNotFound")) {
    lw_lock(&k->lock);
    k->flags |= WORD_NOOK;
    lw_wakeup(&k->lock);
    lw_unlock(&k->lock);
    return;
  }
  if (!strcmp(m->fun, "GameModule_wordSubmitAccepted")) {
    lw_lock(&k->lock);
    k->flags |= WORD_OK;
    lw_wakeup(&k->lock);
    lw_unlock(&k->lock);
    return;
  }
  if (!strcmp(m->fun, "GameModule_userInvitations")) {
    lw_lock(&k->lock);
    k->invit=new_mess("GameModule_userInvitations");
    k->invit->args = m->args;
    k->invit->args_size = m->args_size;
    m->args = 0;
    m->args_size = 0;
    k->flags |= INVITATIONS_OK;
    lw_wakeup(&k->lock);
    lw_unlock(&k->lock);
    return;
  }
  if (verbose) printf("unhandled message %s\n", m->fun);
}

void *net_thread(void *_k)
{
  kelimelik *k = _k;
  int i;
  char *b;
  mess *m;
  while (1) {
    unsigned len = get();
    len = (len << 8) | get();
    len = (len << 8) | get();
    len = (len << 8) | get();
    b = malloc(len); if (!b) abort();
    for (i = 0; i < len; i++) b[i] = get();
    m = mess_unserialize(b, len);
    free(b);
    if (verbose) {
      printf("********************************************************\n");
      print_mess(m);
    }
    handle_mess(k, m);
    free_mess(m);
  }
  return 0;
}

void net_open(kelimelik *k)
{
  struct sockaddr_in addr;

  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock == -1) { perror("socket"); exit(1); }

  addr.sin_family = AF_INET;
  addr.sin_port = htons(k->port);
  addr.sin_addr.s_addr = inet_addr(k->server);

  if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)))
    { perror("connect"); exit(1); }

  new_thread(net_thread, k);
}

void net_send(kelimelik *k, mess *m)
{
  int len;
  char *b = mess_serialize(m, &len);
  if (write(sock, b, len) != len)
    { printf("sock write error\n"); exit(1); }
  free(b);
}
