#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/wait.h>

typedef struct {
  int l;
  int x, y;
} play;

typedef struct {
  int starx, stary;
  int my_score, other_score;
  int remain;
  play po[7];
  int posize;
  play pm[7];
  int pmsize;
  char *hand;
  char *oname;
  char *wname;
  int passline;
} pline;

typedef struct {
  int v[15][15];
  int meplayed;
  int me;
  int other;
  int remain;
  char *hand;
  int points;
} turn;

typedef struct {
  turn *t;
  int tsize;
  int tmaxsize;
  char *winner;
  char *oname;
  int starx, stary;
} game;

int utf8(unsigned char *n, int *_len)
{
  int v = 0;
  if (*n == 0xc0 || *n == 0xc1 ||
      *n >= 0xf5) abort();
  if (!(*n & 0x80)) {
    v = *n;
    *_len = 1;
  } else
  if ((*n & 0xe0) == 0xc0) { /* 110xxxxx */
    if ((*(n+1) & 0xc0) != 0x80) abort();
    v = ((*n & 0x1f) << 6) | (*(n+1) & 0x3f);
    if (v < 0x80 || v > 0x7ff) abort();
    *_len = 2;
  } else
  if ((*n & 0xf0) == 0xe0) { /* 1110xxxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80) abort();
    v = ((*n & 0x0f) << (6*2)) | ((*(n+1) & 0x3f) << 6) | (*(n+2) & 0x3f);
    if (v < 0x800 || v > 0xffff) abort();
    *_len = 3;
  } else
  if ((*n & 0xf8) == 0xf0) { /* 11110xxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80 ||
        (*(n+3) & 0xc0) != 0x80) abort();
    v = ((*n & 0x07) << (6*3)) | ((*(n+1) & 0x3f) << (6*2)) |
        ((*(n+2) & 0x3f) << 6) | (*(n+3) & 0x3f);
    if (v < 0x10000 || v > 0x10ffff) abort();
    *_len = 4;
  } else abort();
  return v;
}

char *toutf8(int c)
{
  static char t[5];
  char *b = t;
  if (c<0x80) *b++=c;
  else if (c<0x800) *b++=192+c/64, *b++=128+c%64;
  else if (c-0xd800u<0x800) abort();
  else if (c<0x10000) *b++=224+c/4096, *b++=128+c/64%64, *b++=128+c%64;
  else if (c<0x110000) *b++=240+c/262144, *b++=128+c/4096%64,
                       *b++=128+c/64%64, *b++=128+c%64;
  else abort();
  *b = 0;
  return t;
}

int utf8_get(FILE *f)
{
  int _;
  unsigned char v[4];
  int c = fgetc(f);
  if (c == EOF) return c;
  if (c < 128) return c;
  v[0] = c;
  if ((c & 0xe0) == 0xc0) {
    v[1] = fgetc(f); if (v[1] == EOF) abort();
    return utf8(v, &_);
  }
  if ((c & 0xf0) == 0xe0) {
    v[1] = fgetc(f); if (v[1] == EOF) abort();
    v[2] = fgetc(f); if (v[2] == EOF) abort();
    return utf8(v, &_);
  }
  if ((c & 0xf8) == 0xf0) {
    v[1] = fgetc(f); if (v[1] == EOF) abort();
    v[2] = fgetc(f); if (v[2] == EOF) abort();
    v[3] = fgetc(f); if (v[3] == EOF) abort();
    return utf8(v, &_);
  }
  abort();
}

void new_turn(game *g)
{
  if (g->tsize == g->tmaxsize) {
    g->tmaxsize += 256;
    g->t = realloc(g->t, g->tmaxsize * sizeof(turn));
    if (!g->t) abort();
  }
  g->t[g->tsize].meplayed = 0;
  g->t[g->tsize].me = 0;
  g->t[g->tsize].other = 0;
  g->t[g->tsize].remain = -1;
  g->t[g->tsize].hand = 0;
  g->tsize++;
}

void dump(int v[15][15], int starx, int stary)
{
  int i, j;
  for (i = 0; i < 15; i++) {
    for (j = 0; j < 15; j++) {
      int x = abs(v[i][j]);
      if (x == '.' && i == stary && j == starx) x = '*';
      printf("%s%s%s", v[i][j]<0?"\033[42;30m":"",toutf8(x),
                       v[i][j]<0?"\033[m":"");
    }
    printf("\n");
  }
}

void load_start(game *g, char *logfile)
{
  char fn[1024]; sprintf(fn, "%s.start", logfile);
  FILE *f = fopen(fn, "r"); if (!f) { perror(fn); exit(1); }
  int i, j;
  new_turn(g);
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    int c;
    do c = utf8_get(f); while (isspace(c));
    g->t[0].v[i][j] = c;
  }
  fclose(f);
}

char *line;
int lsize;
int lmaxsize;

void put(int c)
{
  if (lsize == lmaxsize)
    { lmaxsize += 256; line = realloc(line, lmaxsize); if (!line) abort(); }
  line[lsize++] = c;
}

void get_line(FILE *f)
{
  lsize = 0;
  while (1) {
    int c = fgetc(f); if (c == EOF || c == '\n') break;
    put(c);
  }
  put(0);
}

void parse_log(pline *p, char *l)
{
  int len;
  int i;
  unsigned char *s = (unsigned char *)l;
  char *t = malloc(strlen(l)+1); if (!t) abort();

  p->passline = l[0] == 'P';

  for (i = 0; i < 7; i++) p->po[i].l = p->pm[i].l = 0;
  p->posize = p->pmsize = 0;

  if (sscanf(l, "T %d %d %d %d %d", &p->stary, &p->starx,
             &p->my_score, &p->other_score, &p->remain) != 5 &&
      sscanf(l, "P %d %d %d %d %d", &p->stary, &p->starx,
             &p->my_score, &p->other_score, &p->remain) != 5) abort();

  while (*s && *s != '[') s++; if (!*s) abort(); s++;
  if (*s == '(') goto no_oplay;
  while (*s && *s != ']') {
    int pos=0;
    int c = utf8(s, &len); s+=len;if(*s!=',')abort(); s++;if(!*s)abort();
    while (*s >= '0' && *s <= '9') { pos*=10;pos+=*s-'0';s++; }
    p->po[p->posize].l = c;
    p->po[p->posize].x = pos / 15;
    p->po[p->posize].y = pos % 15;
    if (!(p->po[p->posize].x>=0 && p->po[p->posize].x<15 &&
          p->po[p->posize].y>=0 && p->po[p->posize].y<15)) abort();
    p->posize++;
    if (*s == ']') break; if (*s != ':') abort(); s++;
 }

no_oplay:
  while (*s && *s != '[') s++; if (!*s) abort(); s++;if(!*s)abort();
  if (*s=='(') { p->hand = 0; goto no_hand; }
  len = 0;
  while (*s && *s != ']') { t[len++] = *s; s++; } t[len]=0; if(!len)abort();
  p->hand = strdup(t); if (!p->hand) abort();

no_hand:
  while (*s && *s != ' ') s++; if (!*s)abort(); s++;if(!*s) abort();
  if (sscanf((char*)s, "%d", &len) != 1) abort(); if (!len) abort();
  while (*s && *s!=' ')s++;if(!*s)abort();s++;if(!*s)abort();
  for (i = 0; i < len; i++) {t[i]=*s;s++;if(!*s)abort();} t[i] = 0;
  p->oname = strdup(t); if (!p->oname) abort();

  while (*s && *s != ' ') s++; if (!*s)abort(); s++;if(!*s) abort();
  if (sscanf((char*)s, "%d", &len) != 1) abort();
  if (!len) {
    p->wname=0;
    if (l[0] == 'P') return;
    while (*s && *s != ' ') s++; if (!*s)abort(); s++;if(!*s) abort();
    while (*s && *s == ' ') s++; if (!*s)abort();
    goto no_winner;
  }
  while (*s && *s!=' ')s++;if(!*s)abort();s++;if(!*s)abort();
  for (i = 0; i < len; i++) {if(!*s)abort();t[i]=*s;s++;} t[i] = 0;
  p->wname = strdup(t); if (!p->oname) abort();
  if (l[0] == 'P') return;
  while (*s && *s != ' ') s++; if (!*s) abort();s++;if(!*s)abort();
  while (*s && *s == ' ') s++; if (!*s)abort();

no_winner:
  while (1) {
    if (p->pmsize >= 7) abort();
    p->pm[p->pmsize].l = utf8(s, &len); s+=len; if (!*s) abort();
    p->pmsize++;
    if (*s == ' ') break;
  }
  if (sscanf((char*)s,"%d %d %d",&p->pm[0].y,&p->pm[0].x,&p->pm[1].x)!=3)
    abort();

  free(t);
}

void load_log(game *g, char *logfile)
{
  int i, j, inci, incj, k;
  pline p;
  pline prev; prev.my_score = -1; prev.passline = 1;
  FILE *f = fopen(logfile, "r"); if (!f) { perror(logfile); exit(1); }
  while (1) {
    turn *new;
    get_line(f);
    if (lsize == 1) break;
    parse_log(&p, line);

    g->starx = p.starx; g->stary = p.stary;
    if (!g->oname) g->oname = strdup(p.oname);

    /* old bot had a bug, putting wrong lines in the log, get rid of that */
    if (!prev.passline && p.my_score == prev.my_score) {
      if (g->tsize < 2) abort();
      g->tsize -= 2;
    }

    if (g->tsize) {
      turn *old = &g->t[g->tsize-1];
      old->me = p.my_score;
    }

    if (!g->tsize) goto no_first_other;
    new_turn(g);
    new = &g->t[g->tsize-1];
    memcpy(new->v, g->t[g->tsize-2].v, 15*15*sizeof(int));
    for (i = 0; i < 15; i++) for (j = 0; j < 15; j++)
      new->v[i][j] = abs(new->v[i][j]);
    for (i = 0; i < p.posize; i++)
      new->v[p.po[i].x][p.po[i].y] = -p.po[i].l;
    new->meplayed = 0;

    new->other = p.other_score;
    new->me = p.my_score;
    new->remain = p.remain;
    new->hand = 0;
    if (p.hand) { new->hand = strdup(p.hand); if (!new->hand) abort(); }

no_first_other:
    new_turn(g);
    new = &g->t[g->tsize-1];
    memcpy(new->v, g->t[g->tsize-2].v, 15*15*sizeof(int));
    for (i = 0; i < 15; i++) for (j = 0; j < 15; j++)
      new->v[i][j] = abs(new->v[i][j]);
    i = p.pm[0].y;
    j = p.pm[0].x;
    if (p.pm[1].x) { inci = 1; incj = 0; } else { inci = 0; incj = 1; }
    for (k = 0; k < p.pmsize; k++) {
      while (i>=0 && i<15 && j>=0 && j<15 && new->v[i][j] != '.')
        { i += inci; j += incj; }
      if (!(i>=0 && i<15 && j>=0 && j<15)) abort();
      new->v[i][j] = -p.pm[k].l;
      i+=inci;
      j+=incj;
    }
    new->meplayed = 1;
    new->other = p.other_score;
    new->remain = p.remain - p.pmsize; if (new->remain < 0) new->remain = 0;
    new->hand = 0;
    if (p.hand) { new->hand = strdup(p.hand); if (!new->hand) abort(); }

    free(p.hand); free(p.oname); free(p.wname);
    prev = p;
  }
  fclose(f);
}

void load_over(game *g, char *logfile)
{
  pline p;
  turn *last, *pre;
  char fn[1024]; sprintf(fn, "%s.over", logfile);
  FILE *f = fopen(fn, "r"); if (!f) return;
  int i, j;
  if (!g->tsize) abort();
  new_turn(g);
  last = &g->t[g->tsize-1];
  pre = &g->t[g->tsize-2];
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) {
    int c;
    do c = utf8_get(f); while (isspace(c));
    last->v[i][j] = c;
  }
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++)
    if (last->v[i][j] != abs(pre->v[i][j])) last->v[i][j] *= -1;

  for (i = 0; i < 17; i++) get_line(f);
  lsize = 0;
  put('P'); put(' ');
  while (1) { int c = fgetc(f); if (c == EOF || c == '\n') break; put(c); }
  put(0);
  parse_log(&p, line); /* we don't free, don't care */
  g->starx = p.starx; g->stary = p.stary;
  g->winner = p.wname;
  g->oname = p.oname;

  last->me = p.my_score;
  last->other = p.other_score;
  last->remain = p.remain;
  last->hand = p.hand;

  pre->hand = p.hand;
  pre->remain = p.remain;
  pre->me = p.my_score;

  fclose(f);
}

void compute_points(game *g)
{
  int i;
  g->t[0].points = 0;
  for (i = 1; i < g->tsize; i++) {
    int me = g->t[i].me - g->t[i-1].me;
    int other = g->t[i].other - g->t[i-1].other;
    int points = me > other ? me : other;
    g->t[i].points = points;
  }
}

int utf8strlen(char *s)
{
  if (!s) return 0;
  unsigned char *t = (unsigned char *)s;
  int r = 0;
  while (*t) { if (((*t) & 0xc0) != 0x80) r++; t++; }
  return r;
}

void dump_games(game *g)
{
  char your_score[128];
  char opponent_score[128];
  int other_hand;
  int i, u, v;
  int hh, saved_other_hand = 7;
  for (i = 1; i < g->tsize; i++) {
    turn *t = &g->t[i];
    int me = t->me;
    int other = t->other;
    hh = 0;
    for (u=0;u<15;u++) for (v=0;v<15;v++) {
      if (t->v[u][v] != '.' && !(t->meplayed && t->v[u][v] < 0)) hh++;
    }
    hh += utf8strlen(t->hand);
    other_hand = 98 - hh;
    if (other_hand > saved_other_hand) other_hand = saved_other_hand;
    saved_other_hand = other_hand;
    if (t->meplayed)
      printf("me: ");
    else
      printf("%s: ", g->oname);
    if (me >= other) {
      sprintf(your_score, "\033[42;30m%3d\033[m", me);
      sprintf(opponent_score, "%3d", other);
    } else {
      sprintf(your_score, "%3d", me);
      sprintf(opponent_score, "\033[41;30m%3d\033[m", other);
    }
    printf("%s%s hand [%s] remain %d (%d points) (hh %d)\n",
           your_score, opponent_score,
           t->hand, t->remain, t->points, other_hand);
    dump(t->v, g->starx, g->stary);
  }
}

void usage(void)
{
  printf("replay [options] <game log file>\n");
  printf("options:\n");
  printf("    -gui\n");
  printf("        gui mode\n");
  printf("    -pdf <PDF file\n");
  printf("        pdf mode\n");
  printf("    -scale <scale value (default 0.5)>\n");
  printf("        set the scale value for the gui mode\n");
  exit(1);
}

void guidump(game *g, FILE *f, int pos)
{
  turn *t = &g->t[pos];
  int i, j;
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) if (t->v[i][j] > 0)
    fprintf(f, "%s", toutf8(t->v[i][j])); else fprintf(f, ".");
  fprintf(f, "\n");
  for (i = 0; i < 15; i++) for (j = 0; j < 15; j++) if (t->v[i][j] < 0)
    fprintf(f, "%s", toutf8(-t->v[i][j])); else fprintf(f, ".");
  fprintf(f, "\n");
  fprintf(f, "%s\n%s\n%s\n", t->hand, g->oname, g->winner);
  fprintf(f, "%d %d %d %d %d %d %d %d %d\n", t->me, t->other, t->remain,
          t->points, t->meplayed, g->starx, g->stary, pos, g->tsize);
  fprintf(f, ":\n");
  fflush(f);
}

void do_gui_or_pdf(game *g, char *pdffile, float scale)
{
  FILE *fdad;
  int pos = 0;
  int s;
  int pson[2], pdad[2]; /* 0 read, 1 write */
  if (pipe(pson) || pipe(pdad)) abort();
  s = fork(); if (s==-1) abort();
  if (!s) {
    char sc[64];
    sprintf(sc, "%f", scale);
    close(pson[0]); close(pdad[1]);
    dup2(pson[1], 1);
    dup2(pdad[0], 0);
    if (pdffile) execlp("gui", "gui", "-scale", sc, pdffile, (char*)NULL);
    else         execlp("gui", "gui", "-scale", sc, (char*)NULL);
    fprintf(stderr, "error running 'gui', not in path?\n");
    exit(1);
  }
  close(pdad[0]); close(pson[1]);
  fdad = fdopen(pdad[1], "w"); if (!fdad) abort();

  if (pdffile) {
    int s;
    for (pos = 0; pos < g->tsize - 1; pos++) guidump(g, fdad, pos);
    fprintf(fdad, "+\n"); fflush(fdad);
    wait(&s);
    exit(0);
  }

  guidump(g, fdad, pos);
  while (1) {
    char c;
    if (read(pson[0], &c, 1) != 1) exit(1);
    if (c == 'n' && pos < g->tsize-1) { pos++; guidump(g, fdad, pos); }
    if (c == 'p' && pos)              { pos--; guidump(g, fdad, pos); }
  }
}

game g;
int main(int n, char **v)
{
  char *pdffile = 0;
  char *logfile = 0;
  float scale = .5;
  int i;
  int gui = 0;
  for (i = 1; i < n; i++) {
    if (!strcmp(v[i], "-gui")) { gui = 1; continue; }
    if (!strcmp(v[i], "-pdf")){if(i>n-2)usage();gui=1;pdffile=v[++i];continue;}
    if (!strcmp(v[i], "-scale")){if(i>n-2)usage();scale=atof(v[++i]);continue;}
    if (!logfile) { logfile = v[i]; continue; }
    usage();
  }
  if (!logfile) usage();
  load_start(&g, logfile);
  load_log(&g, logfile);
  load_over(&g, logfile);
  compute_points(&g);
  if (gui) do_gui_or_pdf(&g, pdffile, scale); else dump_games(&g);
  return 0;
}
