#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sndfile.h>
#include <stdarg.h>

#define MAXLEN 48000*10

float buf[MAXLEN];

struct es {
  float freq;
  float t;
  float t60;
  float gain_dB;
  float T;
  float G;
} e[1024];
int esize;

float get(struct es *e, float t)
{
  t -= e->t; if (t < 0) return 0;
  return e->G * exp(t/e->T) * sin(2 * M_PI * e->freq * t);
}

void new_e(float freq, float t, float t60, float gain_dB)
{
  if (esize == 1024) abort();
  e[esize].freq = freq;
  e[esize].t = t;
  e[esize].t60 = t60;
  e[esize].gain_dB = gain_dB;
  e[esize].T = t60 / log(0.001);
  e[esize].G = pow(10, gain_dB / 20);
  esize++;
}

void gen(int len, char *fn, ...)
{
  SF_INFO inf;
  SNDFILE *f;
  int i, j;
  float v;
  va_list ap;

  if (len > MAXLEN) abort();

  esize = 0;

  va_start(ap, fn);
  while (1) {
    float freq = va_arg(ap, double);
    if (freq == 0) break;
    float t = va_arg(ap, double);
    float t60 = va_arg(ap, double);
    float gain_dB = va_arg(ap, double);
    new_e(freq, t, t60, gain_dB);
  }
  va_end(ap);

  inf.samplerate = 48000;
  inf.channels = 1;
  inf.format = SF_FORMAT_WAV | SF_FORMAT_PCM_16 | SF_ENDIAN_LITTLE;
  inf.sections = 1;
  inf.seekable = 1;
  f = sf_open(fn, SFM_WRITE, &inf); if (!f) abort();

  for (i = 0; i < len; i++) {
    v = 0;
    for (j = 0; j < esize; j++) v += get(&e[j], i/48000.);
    buf[i] = v;
  }
  sf_write_float(f, buf, len);

  sf_close(f);
}

int main(void)
{
  gen(12000, "invit.wav",
      800., 0., .15, -20.,
      1200., .08, .15, -30.,
      0.);
  gen(12000, "message.wav",
      600., 0., 0.3, -20.,
      600., .05, 0.3, -26.,
      600., .1, 0.3, -32.,
      0.);
  gen(12000, "pass.wav",
      400., 0., .15, -20.,
      300., .08, .15, -30.,
      0.);
  return 0;
}
