#include "utils.h"
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

void new_rt_thread(void *(*f)(void *), void *data)
{
  pthread_t t;
  pthread_attr_t att;
  int policy;
  struct sched_param sched_params;

  if (pthread_attr_init(&att))
    { fprintf(stderr, "pthread_attr_init err\n"); exit(1); }
  if (pthread_attr_setdetachstate(&att, PTHREAD_CREATE_DETACHED))
    { fprintf(stderr, "pthread_attr_setdetachstate err\n"); exit(1); }
  if (pthread_create(&t, &att, f, data))
    { fprintf(stderr, "pthread_create err\n"); exit(1); }
  if (pthread_attr_destroy(&att))
    { fprintf(stderr, "pthread_attr_destroy err\n"); exit(1); }

  if (pthread_getschedparam(t, &policy, &sched_params))
    perror("pthread_getschedparam");
  if (pthread_setschedprio(t, sched_get_priority_max(policy)))
    perror("pthread_setschedprio");

  sched_params.sched_priority = sched_get_priority_max(SCHED_FIFO);
  if (pthread_setschedparam(t, SCHED_FIFO, &sched_params))
    { perror("pthread_setschedparam");
      fprintf(stderr, "(could not go into SCHED_FIFO)\n"); }
}

void new_thread(void *(*f)(void *), void *data)
{
  pthread_t t;
  pthread_attr_t att;

  if (pthread_attr_init(&att))
    { fprintf(stderr, "pthread_attr_init err\n"); exit(1); }
  if (pthread_attr_setdetachstate(&att, PTHREAD_CREATE_DETACHED))
    { fprintf(stderr, "pthread_attr_setdetachstate err\n"); exit(1); }
  if (pthread_create(&t, &att, f, data))
    { fprintf(stderr, "pthread_create err\n"); exit(1); }
  if (pthread_attr_destroy(&att))
    { fprintf(stderr, "pthread_attr_destroy err\n"); exit(1); }
}

void *new_lock(void)
{
  pthread_mutex_t *m;

  m = malloc(sizeof(pthread_mutex_t)); if (m == NULL) abort();

  if (pthread_mutex_init(m, NULL)) abort();

  return m;
}

void lock(void *l)
{
  if (pthread_mutex_lock(l)) abort();
}

void unlock(void *l)
{
  if (pthread_mutex_unlock(l)) abort();
}

void *new_cond(void)
{
  pthread_cond_t *c;

  c = malloc(sizeof(pthread_cond_t)); if (c == NULL) abort();

  if (pthread_cond_init(c, NULL)) abort();

  return c;
}

void cond_wait(void *cond, void *lock)
{
  if (pthread_cond_wait(cond, lock)) abort();
}

void cond_wakeup(void *cond)
{
  if (pthread_cond_broadcast(cond)) abort();
}

void lw_init(lockwait *l)
{
  l->mutex = new_lock();
  l->cond = new_cond();
}

void lw_lock(lockwait *l)
{
  lock(l->mutex);
}

void lw_unlock(lockwait *l)
{
  unlock(l->mutex);
}

void lw_wait(lockwait *l)
{
  cond_wait(l->cond, l->mutex);
}

void lw_wakeup(lockwait *l)
{
  cond_wakeup(l->cond);
}

void init_queue(queue *q)
{
  lw_init(&q->l);
  q->head = q->tail = NULL;
}

void *wait_dequeue(queue *q)
{
  void *r, *f;
  lw_lock(&q->l);
  while (q->tail == NULL) lw_wait(&q->l);
  r = q->tail->data;
  f = q->tail;
  q->tail = q->tail->prev;
  if (q->tail == NULL) q->head = NULL;
  free(f);
  lw_unlock(&q->l);
  return r;
}

void enqueue(queue *q, void *d)
{
  queue_item *new;
  lw_lock(&q->l);
  new = malloc(sizeof(queue_item)); if (new == NULL) abort();
  new->data = d;
  new->prev = NULL;
  if (q->head == NULL) q->head = q->tail = new;
  else                 q->head->prev = new;
  lw_wakeup(&q->l);
  lw_unlock(&q->l);
}
